/*
 * Access a PWM server to update a display panel backlight.
 *
 * Copyright (C) 2018, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env>
#include <l4/sys/capability>

#include <stdlib.h>

#include <l4/devices/pwm-client.h>

#include <ipc/server.h>
#include "backlight_object_server.h"

/* PWM access abstractions. */

static L4::Cap<Pwm_device_interface> pwm_device;



/* Backlight device. */

class server_BacklightObject : public BacklightObject
{
  int _min, _max;

public:
  explicit server_BacklightObject(int min, int max)
  : _min(min), _max(max)
  {
  }

  /* Disable the backlight. */

  long disable()
  {
    pwm_device->disable();
    return L4_EOK;
  }

  /* Enable the backlight. */

  long enable()
  {
    pwm_device->enable();
    return L4_EOK;
  }

  /* Use the PWM device to update the backlight level. */

  long set_brightness(int level)
  {
    pwm_device->set_duty(level);
    return L4_EOK;
  }
};



/* Arguments: <minimum level> <maximum level> */

int main(int argc, char *argv[])
{
  int min, max;

  if (argc < 3)
    return 1;

  /* Obtain a reference to the PWM device. */

  pwm_device = L4Re::Env::env()->get_cap<Pwm_device_interface>("pwm");
  if (!pwm_device.is_valid()) return 1;

  /* Initialise and register a new server object. */

  min = atoi(argv[1]);
  max = atoi(argv[2]);

  server_BacklightObject obj(min, max);
  l4_cap_idx_t server;

  if (ipc_server_bind("backlight", (l4_umword_t) &obj, &server)) return 1;

  /* Enter the IPC server loop. */

  ipc_server_loop(BacklightObject_expected_items, &obj,
                  (ipc_server_handler_type) handle_BacklightObject);
  return 0;
}
