/*
 * LCD peripheral support for the JZ4740 and related SoCs.
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017, 2018,
 *               2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include <l4/sys/cache.h>
#include <l4/sys/types.h>

#include "lcd-jz4780.h"
#include "lcd-jz4740-config.h"
#include "lcd-jz4740-regs.h"

#include <stdint.h>



// JZ4780-specific methods.

Lcd_jz4780_chip::Lcd_jz4780_chip(l4_addr_t addr, Jz4740_lcd_panel *panel)
: Lcd_jz4740_chip(addr, panel)
{
  _burst_size = 64; // 64-word burst size available in the JZ4780
}

void
Lcd_jz4780_chip::enable()
{
  _regs[Osd_status] = 0;
  Lcd_jz4740_chip::enable();
}

// Return an interrupt-related OSD configuration value.

uint32_t
Lcd_jz4780_chip::_osd_config_irq()
{
  return ((_irq_conditions & Lcd_irq_frame_start) ? (1U << Osd_config_fg0_frame_start_irq_enable) : 0) |
         ((_irq_conditions & Lcd_irq_frame_end) ? (1U << Osd_config_fg0_frame_end_irq_enable) : 0);
}

// Return colour depth control value.
// JZ4780 position details only.

uint32_t
Lcd_jz4780_chip::_position_bpp()
{
  uint32_t value;

  switch (_panel->bpp)
  {
    case 15: case 16: value = Position_bpp_15_16bpp; break;
    case 18: case 24: value = Position_bpp_18_24bpp; break;
    case 30: value = Position_bpp_30bpp; break;
    default: value = 0; break;
  }

  return value << Position_bpp;
}

uint32_t
Lcd_jz4780_chip::_priority_transfer()
{
  uint32_t length;

  switch (_burst_size)
  {
    case 4:  length = Priority_burst_4;  break;
    case 8:  length = Priority_burst_8;  break;
    case 32: length = Priority_burst_32; break;
    case 64: length = Priority_burst_64; break;
    case 16:
    default: length = Priority_burst_16; break;
  }

  return Priority_mode_arbiter |
         (length << Priority_highest_burst) |
         (511U << Priority_threshold2) |
         (400U << Priority_threshold1) |
         (256U << Priority_threshold0);
}

// Initialise a DMA descriptor for the JZ4780. The principal differences with
// earlier SoCs are the "new" descriptor fields which populate additional
// registers controlling OSD foreground planes, and the frame enable flag which
// allows the descriptors/planes to be disabled and left unused.

void
Lcd_jz4780_chip::_set_descriptor(struct Jz4740_lcd_descriptor &desc,
                                 l4_addr_t source, l4_size_t size,
                                 struct Jz4740_lcd_descriptor *next,
                                 uint32_t flags,
                                 bool frame_enable)
{
  // In the command, indicate the number of words from the source for transfer.

  desc.next = next;
  desc.source = frame_enable ? source : 0;
  desc.identifier = source;
  desc.command = ((size / sizeof(uint32_t)) & Command_buffer_length_mask) |
                 (frame_enable ? (1U << Command_frame_enable) : 0) |
                 flags;

  // Initialise "new" descriptor fields.

  desc.offset = 0;
  desc.page_width = 0;
  desc.command_position = (1U << Position_premultiply_lcd) |
                          ((frame_enable ? 1U : 3U) << Position_coefficient) |
                          _position_bpp();
  desc.fg_size = 0xff000000 |
                 ((_panel->height - 1) << 12) |
                 ((_panel->width - 1) << 0);
}

// HDMI-compatible JZ4780 configuration.
// Remarks in the Ingenic Linux 3.0.8 driver suggest that the JZ4775 and JZ4780
// do not support palettes. Here, multiple panels are also not supported.

void
Lcd_jz4780_chip::config(struct Jz4740_lcd_descriptor *desc_vaddr,
                        struct Jz4740_lcd_descriptor *desc_paddr,
                        l4_addr_t fb_paddr)
{
  // Provide the first framebuffer descriptor in single and dual modes.
  // Flip back and forth between any palette and the framebuffer.

  _set_descriptor(desc_vaddr[0], get_framebuffer(0, fb_paddr),
                                 get_aligned_size(),
                                 desc_paddr,
                                 _command_irq());

  // Provide the second framebuffer descriptor only in dual-panel mode.
  // Only employ this descriptor in the second DMA channel.

  _set_descriptor(desc_vaddr[1], get_framebuffer(1, fb_paddr),
                                 get_aligned_size(),
                                 desc_paddr + 1,
                                 _command_irq(),
                                 false);

  // Flush cached structure data.

  l4_cache_clean_data((unsigned long) desc_vaddr,
                      (unsigned long) desc_vaddr + get_descriptors_size());

  // Configure DMA by setting frame descriptor addresses.

  // Provide the palette descriptor address first, if employed.

  _regs[Desc_address_0] = (uint32_t) desc_paddr;

  // Provide a descriptor for the second DMA channel, providing foreground 1.

  _regs[Desc_address_1] = (uint32_t) (desc_paddr + 1);

  // Initialise panel-related registers.

  _init_panel();

  // Initialise the control and configuration registers.
  // JZ4780 does not support bpp setting here. Otherwise, this is the same as
  // with earlier SoCs.

  _regs[Lcd_control] = _control_panel() | _control_transfer() | _control_irq();
  _regs[Lcd_config] = _panel->config;

  // JZ4780-specific configuration.
  // The RGB control register usage may be superfluous.

  _regs[Rgb_control] = (1U << Rgb_format_enable) | Rgb_odd_line_rgb | Rgb_even_line_rgb;
  _regs[Priority_level] = _priority_transfer();

  // Employ whole image alpha levels by default.

  _regs[Osd_config] = (1U << Osd_config_enable) |
                      (1U << Osd_config_alpha_enable);
  _regs[Alpha_levels] = ((255U << Alpha_level_fg1) & Alpha_level_fg1_mask) |
                        ((255U << Alpha_level_fg0) & Alpha_level_fg0_mask);
}



// C language interface functions.

void *
jz4780_lcd_init(l4_addr_t lcd_base, struct Jz4740_lcd_panel *panel)
{
  return (void *) new Lcd_jz4780_chip(lcd_base, panel);
}
